use super::Stream;
use crate::{
	local_socket::{prelude::*, traits::tokio as traits, ListenerNonblockingMode, ListenerOptions},
	os::unix::uds_local_socket::{listener::Listener as SyncListener, ReclaimGuard},
	Sealed,
};
use futures_core::{FusedStream as FusedAsyncIterator, Stream as AsyncIterator};
use std::{
	fmt::{self, Debug, Formatter},
	future::Future,
	io,
	os::unix::prelude::*,
	pin::{pin, Pin},
	task::{Context, Poll},
};
use tokio::net::UnixListener;

pub struct Listener {
	listener: UnixListener,
	reclaim: ReclaimGuard,
}
impl Sealed for Listener {}
impl traits::Listener for Listener {
	type Stream = Stream;

	fn from_options(options: ListenerOptions<'_>) -> io::Result<Self> {
		options
			.nonblocking(ListenerNonblockingMode::Neither)
			.create_sync_as::<SyncListener>()
			.and_then(Self::try_from)
	}
	async fn accept(&self) -> io::Result<Stream> {
		let inner = self.listener.accept().await?.0;
		Ok(Stream::from(inner))
	}

	fn do_not_reclaim_name_on_drop(&mut self) {
		self.reclaim.forget();
	}
}
impl AsyncIterator for Listener {
	type Item = io::Result<Stream>;
	#[inline(always)]
	fn poll_next(self: Pin<&mut Self>, cx: &mut Context<'_>) -> Poll<Option<Self::Item>> {
		pin!(traits::Listener::accept(self.get_mut()))
			.poll(cx)
			.map(Some)
	}
}
impl FusedAsyncIterator for Listener {
	#[inline(always)]
	fn is_terminated(&self) -> bool {
		false
	}
}

impl TryFrom<SyncListener> for Listener {
	type Error = io::Error;
	fn try_from(mut sync: SyncListener) -> io::Result<Self> {
		sync.set_nonblocking(ListenerNonblockingMode::Accept)?;
		let reclaim = sync.reclaim.take();
		Ok(Self {
			listener: UnixListener::from_std(sync.into())?,
			reclaim,
		})
	}
}

impl Debug for Listener {
	fn fmt(&self, f: &mut Formatter<'_>) -> fmt::Result {
		f.debug_struct("Listener")
			.field("fd", &self.listener.as_raw_fd())
			.field("reclaim", &self.reclaim)
			.finish()
	}
}
impl AsFd for Listener {
	#[inline]
	fn as_fd(&self) -> BorrowedFd<'_> {
		self.listener.as_fd()
	}
}
impl TryFrom<Listener> for OwnedFd {
	type Error = io::Error;
	fn try_from(mut slf: Listener) -> io::Result<Self> {
		slf.listener.into_std().map(|s| {
			slf.reclaim.forget();
			s.into()
		})
	}
}
impl TryFrom<OwnedFd> for Listener {
	type Error = io::Error;
	fn try_from(fd: OwnedFd) -> io::Result<Self> {
		Self::try_from(SyncListener::from(fd))
	}
}
